@@ -1,5 +1,5 @@
 /*
- renattach 1.2.1 - Filter that renames/deletes dangerous email attachments
+ renattach 1.2.2 - Filter that renames/deletes dangerous email attachments
  Copyright (C) 2003, 2004  Jem E. Berkes <jberkes@pc-tools.net>
 
  This program is free software; you can redistribute it and/or modify
@@ -34,9 +34,24 @@
 #include <signal.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <syslog.h>
+#include <sys/wait.h>
 #include <unistd.h>
 
+#ifdef HAVE_SYSLOG_H
+#include <syslog.h>
+#endif
+
+#ifdef __EMX__
+#include <io.h>
+#endif
+
+#ifndef HAVE_STRCASECMP
+#define strcasecmp stricmp
+#endif
+
+#ifndef HAVE_STRNCASECMP
+#define strncasecmp strnicmp
+#endif
 
 /* Function prototypes */
 int filter_pass(FILE*);
@@ -46,12 +61,14 @@ int filter_decision(struct attach*);
 int match_filename(const char*, int);
 int decode_2047(char*, struct namespec*);
 int reencode_name(struct namespec*, const char*, char*);
+int position_html(FILE*, FILE*);
 void kill_exit(const char*);
 void taking_action(const char*);
 void not_taking_action(const char*, const char*);
 int explain_subject(FILE*, char*);
-void sanitize_shell_cmd(char*);
-void abnormalproc(void);
+void abnormalproc(int);
+FILE* openpipe(char**, pid_t*);
+int closepipe(FILE*, pid_t, int*);
 
 
 /* Global variables */
@@ -172,23 +189,15 @@ int main(int argc, char** argv)
 				break;
 
 			case 'p':
-				if (*optarg == '\"')	/* caller didn't interpret quotes */
-				{
-					/* Build long option from all following argv's */
-					expand_list(&configuration.pipe_cmd, optarg+1, " ");
-					for (; optind < argc; optind++)
-						expand_list(&configuration.pipe_cmd, argv[optind], " ");
-					/* Since we removed opening quotes, remove closing too */
-					if (configuration.pipe_cmd[strlen(configuration.pipe_cmd)-1] == '\"')
-						configuration.pipe_cmd[strlen(configuration.pipe_cmd)-1] = '\0';
-				}
-				else
-				{
-					configuration.pipe_cmd = malloc(1+strlen(optarg));
-					strcpy(configuration.pipe_cmd, optarg);
-				}
-				break;
-
+			{
+				/* The pipe command is everything else on the command line, as is */
+				int argnum = 0;
+				configuration.pipe_cmd = (char**)calloc(argc, sizeof(char*));
+				for (optind--; optind<argc; optind++)
+					configuration.pipe_cmd[argnum++] = argv[optind];
+				break;			
+			}
+			
 			case 's':
 				justsettings = 1;
 				break;
@@ -217,8 +226,6 @@ int main(int argc, char** argv)
 		configuration.mode = MODE_BAD;		/* default filter mode is badlist */
 	if (!configuration.def_act)
 		configuration.def_act = ACTION_RENAME;	/* default filter action is to rename */
-	if (configuration.pipe_cmd)
-		sanitize_shell_cmd(configuration.pipe_cmd);
 
 	/* Start filter with current action = default action */
 	configuration.action = configuration.def_act;
@@ -500,6 +507,7 @@ int filter_pass(FILE* dest)
 	This header modification pass doesn't do any actual filtering.
 	The previously filtered mail is read back from source, and some headers
 	are added and modified before writing to destination (stdout or pipe).
+	A warning message (warning_text and/or warning_html) may also be added.
 	source is file handle to temporary file
 	Returns nonzero on success, zero on failure
 */
@@ -509,7 +517,15 @@ int header_pass(FILE* source)
 	long endpos;
 	int onheaders = 1;
 	int subject_changed = 0;
+	int plain_part = 0;
+	int html_part = 0;
+	int plain_warning = (configuration.warning_text && *configuration.warning_text);
+	int html_warning = (configuration.warning_html && *configuration.warning_html);
 	FILE* dest = stdout;
+	pid_t pid;
+	
+	if (plain_warning)
+		plain_part = 1;
 
 	/* Stream data may have been overwritten, so we need to set new end of file */
 	endpos = ftell(source);
@@ -524,10 +540,10 @@ int header_pass(FILE* source)
 	/* Open output destination pipe, if requested */
 	if (configuration.pipe_cmd)
 	{
-		dest = popen(configuration.pipe_cmd, "w");
+		dest = openpipe(configuration.pipe_cmd, &pid);
 		if (dest == NULL)
 		{
-			perror("popen");
+			fprintf(stderr, TXT_ERR_OPENPIPE ": %s\n", configuration.pipe_cmd[0]);
 			return 0;	/* error opening pipe */
 		}
 		signal(SIGPIPE, abnormalproc);	/* establish signal handler for pipe error */
@@ -535,43 +551,79 @@ int header_pass(FILE* source)
 
 	while (fgets(linebuf, sizeof(linebuf), source))
 	{
-		/* Modify subject if required */
-		if (onheaders && filtcount && (strncasecmp(linebuf, HEAD_SUBJECT, sizeof(HEAD_SUBJECT)-1)==0))
+		/* Reached blank line */
+		if ((*linebuf == '\n') || (*linebuf == '\r'))
 		{
-			subject_changed = 1;
-			if (explain_subject(dest, linebuf+sizeof(HEAD_SUBJECT)-1))
-				continue;
-		}
-
-		/* Keep existing X- headers specific to RenAttach, by renaming as Old- */
-		if (onheaders && ( (strncasecmp(linebuf, HEAD_FILTERED_WITH, sizeof(HEAD_FILTERED_WITH)-1)==0)
-			|| (strncasecmp(linebuf, HEAD_FILTERED_INFO, sizeof(HEAD_FILTERED_INFO)-1)==0) ))
-		{
-			fprintf(dest, "Old-%s", linebuf);
-			continue;
-		}
-
-		/* Append our new headers */
-		if (onheaders && ((*linebuf == '\n') || (*linebuf == '\r')) )
-		{
-			onheaders = 0;
-			fprintf(dest, "X-Filtered-With: %s\n", PACKAGE_STRING);
-			fprintf(dest, "X-RenAttach-Info: "TXT_HEAD_MODE"=%s " TXT_HEAD_ACTION"=%s " \
-				TXT_HEAD_FILTCOUNT"=%d", mode2str(configuration.mode),
-				act2str(configuration.action), filtcount);
-			if (filtcount)
+			/* If we were still on headers, then append our new headers */
+			if (onheaders)
 			{
-				fprintf(dest, " (%s)\n", TXT_HEAD_FILTERED);
-				if (!subject_changed)
-					explain_subject(dest, NULL);	/* add new Subject */
+				onheaders = 0;
+				fprintf(dest, "X-Filtered-With: %s\n", PACKAGE_STRING);
+				fprintf(dest, "X-RenAttach-Info: "TXT_HEAD_MODE"=%s " TXT_HEAD_ACTION"=%s " \
+					TXT_HEAD_FILTCOUNT"=%d", mode2str(configuration.mode),
+					act2str(configuration.action), filtcount);
+				if (filtcount)
+				{
+					fprintf(dest, " (%s)\n", TXT_HEAD_FILTERED);
+					if (!subject_changed)
+						explain_subject(dest, NULL);	/* add new Subject */
+				}
+				else
+					fprintf(dest, "\n");
+				/* If filtering occurred and there are custom headers, add them */
+				if (filtcount && configuration.add_header && *configuration.add_header)
+					fprintf(dest, "%s\n", configuration.add_header);
 			}
-			else
+			
+			/* We've reached start of body part, insert warning if desired */
+			if (filtcount && plain_part)
+			{
+				fprintf(dest, "\n%s\n", configuration.warning_text);
+				plain_part = 0;
+			}
+			else if (filtcount && html_part)
+			{
 				fprintf(dest, "\n");
-			/* If filtering occurred and there are custom headers, add them */
-			if (filtcount && configuration.add_header && *configuration.add_header)
-				fprintf(dest, "%s\n", configuration.add_header);
+				if (position_html(source, dest) != 0)
+					return 0;	/* error during seeking */
+				fprintf(dest, "%s", configuration.warning_html);
+				fgets(linebuf, sizeof(linebuf), source);	/* because source shifted */
+				html_part = 0;
+			}
 		}
-
+		
+		/* Processing of message headers */
+		if (onheaders)
+		{
+			/* Modify subject if required */
+			if (filtcount && (strncasecmp(linebuf, HEAD_SUBJECT, sizeof(HEAD_SUBJECT)-1)==0))
+			{
+				subject_changed = 1;
+				if (explain_subject(dest, linebuf+sizeof(HEAD_SUBJECT)-1))
+					continue;
+			}
+	
+			/* Keep existing X- headers specific to RenAttach, by renaming as Old- */
+			if ( (strncasecmp(linebuf, HEAD_FILTERED_WITH, sizeof(HEAD_FILTERED_WITH)-1)==0)
+				|| (strncasecmp(linebuf, HEAD_FILTERED_INFO, sizeof(HEAD_FILTERED_INFO)-1)==0) )
+			{
+				fprintf(dest, "Old-%s", linebuf);
+				continue;
+			}
+		}
+		
+		/* If warning_text or warning_html is enabled, search for next parts */
+		if (plain_warning && (strncasecmp(linebuf, MIME_TYPETEXT, sizeof(MIME_TYPETEXT)-1) == 0))
+		{
+			plain_part = 1;
+			html_part = 0;
+		}
+		else if (html_warning && (strncasecmp(linebuf, MIME_TYPEHTML, sizeof(MIME_TYPEHTML)-1) == 0))
+		{
+			plain_part = 0;
+			html_part = 1;
+		}
+		
 		if (fputs(linebuf, dest) == EOF)
 			return 0;	/* fputs error */
 	}
@@ -579,11 +631,13 @@ int header_pass(FILE* source)
 	/* Make sure output gets written */
 	if (dest != stdout)
 	{
-		int retcode = pclose(dest);
+		int retcode = -1;
+		if (closepipe(dest, pid, &retcode) != 0)
+			fprintf(stderr, TXT_ERR_CLOSEPIPE ": %s\n", configuration.pipe_cmd[0]);
 		if (verbose)
-			fprintf(stderr, "%s: %d\n", TXT_INFO_PIPEXITCODE, retcode);
+			fprintf(stderr, "%s: 0x%X\n", TXT_INFO_PIPEXITCODE, retcode);
 		if (retcode != 0)	/* by convention, nonzero exitcode indicates error condition */
-			return 0;	/* header pass failed due to pipe error*/
+			return 0;	/* header pass failed due to pipe error */
 	}
 	return 1;
 }
@@ -688,7 +742,7 @@ int is_line_mime(const char* line, int on_mime, struct attach* m_attachment)
 				line += strlen(parsed);	/* further advance pointer for next search */
 				if (m_attachment->curspec->specformat == FMT_RFC2231)
 					decode_hex(parsed, '%', 0, NULL);
-				else if ( strstr(parsed, "=?") && (decode_2047(parsed, m_attachment->curspec)==0) )
+				else if ((strstr(parsed, "=?") || strstr(parsed, "?=")) && (decode_2047(parsed, m_attachment->curspec)==0) )
 					m_attachment->curspec->specformat = FMT_RFC2047;
 				else
 				{
@@ -959,6 +1013,7 @@ int decode_2047(char* field, struct namespec* fieldspec)
 	int pos;	/* position marker */
 	int field_len;	/* length of field */
 	int improper;	/* flag set if improper coding found */
+	int valid = 0;	/* flag set when valid coding found */
 	int state = 0;	/*	0 = normal
 				1 = just inside encoded word
 				2 = found pre-encoding ? */
@@ -981,8 +1036,10 @@ int decode_2047(char* field, struct namespec* fieldspec)
 		{
 			case 0:
 				/* Make a note of the character encoding */
-				if (sscanf(field+pos, scanspec, fieldspec->charenc) == 1)
+				if (	(sscanf(field+pos, scanspec, fieldspec->charenc) == 1) ||
+					(sscanf(field+pos, scanspec+1, fieldspec->charenc) == 1) )
 				{
+					valid = 1;
 					state++;
 					pos++;
 				}
@@ -1043,7 +1100,7 @@ int decode_2047(char* field, struct namespec* fieldspec)
 		}
 	}
 	
-	if (improper)
+	if (improper || !valid)
 	{
 		free(result);
 		*field = 0;	/* clear the invalid field */
@@ -1201,6 +1258,64 @@ int reencode_name(struct namespec* fieldspec, const char* tagbase, char* result)
 }
 
 
+/*
+	This function finds the "right place" (based on htmlwarn_pos from .conf file) to insert warning_html.
+	The current source_pos is recorded, then source is read through in search of all htmlwarn_pos tags.
+	Once the source_htmlspot is found, the region from source_pos to source_htmlpos is written to dest.
+	When the function returns, dest points to the correct spot to output the warning_html.
+	Returns 0 on success, nonzero if seeking fails
+*/
+int position_html(FILE* source, FILE* dest)
+{
+	char* tag = NULL;
+	char linebuf[MAXFIELD];
+	char taglist[MAXFIELD];
+	long dump_size;
+	long source_start, source_htmlspot;
+
+	source_start = ftell(source);
+	if (source_start == -1)
+		return -1;
+	source_htmlspot = source_start;
+	strcpy(taglist, configuration.htmlwarn_pos);
+	tag = strtok(taglist, LIST_TOKENS);
+	while (tag && fgets(linebuf, sizeof(linebuf), source))
+	{
+		char *tagstart, *tagend;
+		char fulltag[MAXFIELD];
+		sprintf(fulltag, "<%s", tag);
+		if ((tagstart = stristr(linebuf, fulltag)) && (tagend = strchr(tagstart, '>')))
+		{
+			/* Located one more tag, note the best spot found so far */
+			source_htmlspot = ftell(source);
+			if (source_htmlspot == -1)
+				return -1;
+			source_htmlspot = source_htmlspot - strlen(linebuf) + (tagend - linebuf) + 1;
+			if (fseek(source, source_htmlspot, SEEK_SET) != 0)	/* backtrack for next fgets */
+				return -1;
+			tag = strtok(NULL, LIST_TOKENS);
+		}
+		if (strncasecmp(linebuf, MIME_CTYPE, sizeof(MIME_CTYPE)-1) == 0)
+			break;
+	}
+	/* Either advanced through all tags (ideal) or reached end of source */
+	if (fseek(source, source_start, SEEK_SET) != 0)
+		return -1;
+	dump_size = source_htmlspot - source_start;
+	if (dump_size > 0)
+	{
+		char* dump_buf = calloc(1, 1+dump_size);
+		if (!dump_buf)
+			return -1;
+		if (fread(dump_buf, dump_size, 1, source) < 1)
+			return -1;
+		fprintf(dest, "%s", dump_buf);
+		free(dump_buf);
+	}
+	return 0;
+}
+
+
 /*
 	Kill this email and exit the program, taking_action along the way
 */
@@ -1227,9 +1342,11 @@ void taking_action(const char* caughtfile)
 		fprintf(stderr, "%s \"%s\" in %s", act2str(configuration.action), caughtfile, messageid);
 	if (configuration.use_syslog == OPTION_ENABLED)
 	{
+		#ifdef HAVE_SYSLOG_H
 		openlog("renattach", LOG_PID, LOG_MAIL);
 		syslog(LOG_WARNING, "%s \"%s\" in %s", act2str(configuration.action), caughtfile, messageid);
 		closelog();
+		#endif
 	}
 	/* Update filter counters */
 	filtcount++;
@@ -1252,9 +1369,11 @@ void not_taking_action(const char* reason, const char* missedfile)
 		fprintf(stderr, TXT_INFO_NOTACTING " - %s: %s\n", reason, missedfile);
 	if (configuration.use_syslog == OPTION_ENABLED)
 	{
+		#ifdef HAVE_SYSLOG_H
 		openlog("renattach", LOG_PID, LOG_MAIL);
 		syslog(LOG_WARNING, TXT_INFO_NOTACTING " - %s: %s", reason, missedfile);
 		closelog();
+		#endif
 	}
 	/* Reset to default action */
 	configuration.action = configuration.def_act;
@@ -1316,43 +1435,83 @@ int explain_subject(FILE* dest, char* oldsubject)
 }
 
 
-/*
-	Shell commands are sanitized by stripping out dangerous characters
-*/
-void sanitize_shell_cmd(char* buf)
-{
-	int inpos, outpos=0, length=strlen(buf);
-	char* tmpbuf = calloc(1, 1+length);
-	for (inpos=0; inpos<length; inpos++)
-	{
-		switch (buf[inpos])
-		{
-			case SEMICOLON:
-			case BACKTICK:
-			case AMPERSAND:
-			case DOLLARSIGN:
-			case BACKSLASH:
-			case GREATERTHAN:
-			case LESSTHAN:
-			case VERTICALBAR:
-				break;
-
-			default:
-				tmpbuf[outpos++] = buf[inpos];
-		}
-	}
-	strcpy(buf, tmpbuf);
-	free(tmpbuf);
-}
-
-
 /*
 	SIGPIPE signal handler; clean up tempfile, exit
 */
-void abnormalproc()
+void abnormalproc(int sig)
 {
 	if (tempfile)
 		fclose(tempfile);
 	fprintf(stderr, TXT_INFO_SIGPIPE "\n");
 	exit(CODE_TEMPFAIL);	/* temporary failure */
 }
+
+
+/*
+	Opens a write-pipe to indicated command, where args[0] is program
+	childpid receives the child's process ID, required for closepipe()	
+	
+	Returns FILE* for writing to on success
+	Returns NULL if pipe/fork/exec fails, or program terminates prematurely
+*/
+FILE* openpipe(char** args, pid_t* childpid)
+{
+	FILE* command;
+	int cmdpipe[2];
+
+	if (pipe(cmdpipe) != 0)
+		return NULL;
+	switch (*childpid = fork())
+	{
+		case -1:	/* error */
+			close(cmdpipe[0]);
+			close(cmdpipe[1]);
+			return NULL;
+
+		case 0:		/* child */
+			if (dup2(cmdpipe[0], STDIN_FILENO) == -1)
+				_exit(-1);
+			close(cmdpipe[0]);
+			close(cmdpipe[1]);
+			execv(args[0], args);
+			fprintf(stderr, TXT_ERR_EXEC ": %s\n", args[0]);
+			_exit(-1);
+	
+		default:	/* parent */
+			close(cmdpipe[0]);			/* close unused pipe end */
+			command = fdopen(cmdpipe[1], "w");	/* open write-side of pipe */
+			if (command)
+				return command;
+			else
+			{
+				close(cmdpipe[1]);
+				return NULL;
+			}
+	}
+}
+
+
+/*
+	Close the write-pipe opened earlier with openpipe()
+	
+	Returns 0 on normal termination, exitcode filled in
+	Otherwise returns nonzero value, exitcode unmodified
+*/
+int closepipe(FILE* command, pid_t childpid, int* exitcode)
+{
+	int status;
+	if (fclose(command) != 0)
+	{
+		waitpid(childpid, &status, 0);	/* cleanup any way */
+		return -1;		/* error closing stream/pipe */
+	}
+	if (waitpid(childpid, &status, 0) == -1)
+		return -1;		/* error waiting for child */
+	if (WIFEXITED(status))
+	{
+		*exitcode = WEXITSTATUS(status);
+		return 0;		/* terminated normally */
+	}
+	else
+		return -1;		/* abnormal termination or nonzero exit code */
+}
